/*----------------------------------------------------------------------
  File    : nposs.h
  Contents: Naive possibilistic classifier management
  Author  : Christian Borgelt
  History : 2001.02.07 file created from file nbayes.h
            2001.07.17 adapted to modified module scan
            2004.08.12 adapted to new module parse (removed later)
            2011.12.09 adapted to modified module scanner
            2013.08.07 adapted to definitions ATTID, VALID, TPLID etc.
----------------------------------------------------------------------*/
#ifndef __NPOSS__
#define __NPOSS__
#ifdef NPC_PARSE
#ifndef SCN_SCAN
#define SCN_SCAN
#endif
#include "scanner.h"
#endif
#include "table.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- induction modes --- */
#define NPC_CLONE   0x0002      /* clone attribute set */
#define NPC_ADD     0x0004      /* greedily add attributes */
#define NPC_REMOVE  0x0008      /* greedily remove attributes */

/* --- description modes --- */
#define NPC_TITLE   0x0001      /* print a title (as a comment) */
#define NPC_INFO    0x0002      /* print add. info. (as a comment) */
#define NPC_POSS    0x0004      /* print degrees of possibility */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- possibility distribution --- */
  double max;                   /* maximal frequency */
  double *frqs;                 /* value frequency   vector */
  double *poss;                 /* value possibility vector */
} POSSD;                        /* (possibility distribution) */

typedef struct {                /* --- distribution vector --- */
  ATTID  mark;                  /* whether read or to be processed */
  int    type;                  /* attribute type (0: class) */
  VALID  cnt;                   /* number of attribute values */
  POSSD  *possds;               /* vector of poss. distributions */
} CDIST;                        /* (conditional distribution) */

typedef struct {                /* --- naive poss. classifier --- */
  ATTSET *attset;               /* underlying attribute set */
  ATTID  attcnt;                /* number of attributes */
  ATT    *clsatt;               /* class attribute */
  ATTID  clsid;                 /* identifier of class attribute */
  VALID  clscnt;                /* number of classes */
  double total;                 /* total number of cases */
  double max;                   /* maximum class possibility */
  double *frqs;                 /* class frequency vector */
  double *priors;               /* prior     class possibilities */
  double *posts;                /* posterior class possibilities */
  double *buf;                  /* buffer (for internal use) */
  CDIST  dists[1];              /* conditional distributions */
} NPC;                          /* (naive possibilistic classifier) */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/
extern NPC*    npc_create (ATTSET *attset, ATTID clsid);
extern NPC*    npc_clone  (NPC *npc, int cloneas);
extern void    npc_delete (NPC *npc, int delas);
extern void    npc_clear  (NPC *npc);

extern ATTSET* npc_attset (const NPC *npc);
extern ATTID   npc_attcnt (const NPC *npc);
extern VALID   npc_valcnt (const NPC *npc, ATTID attid);
extern ATT*    npc_clsatt (NPC *npc);
extern ATTID   npc_clsid  (const NPC *npc);
extern VALID   npc_clscnt (const NPC *npc);
extern double  npc_total  (const NPC *npc);

#ifdef NPC_INDUCE
extern int     npc_add    (NPC *npc, const TUPLE *tpl);
extern NPC*    npc_induce (TABLE *table, ATTID clsid, int mode);
extern ATTID   npc_mark   (NPC *npc);
#endif

extern void    npc_setup  (NPC *npc);
extern double  npc_prior  (const NPC *npc, VALID clsid);
extern double  npc_poss   (const NPC *npc, VALID clsid, ATTID attid,
                           VALID valid);
extern double  npc_post   (const NPC *nbc, VALID clsid);
extern VALID   npc_exec   (const NPC *npc, const TUPLE *tpl,
                           double *conf);

extern int     npc_desc   (const NPC *npc, FILE *file, int mode,
                           int maxlen);
#ifdef NPC_PARSE
extern NPC*    npc_parse  (ATTSET *attset, SCANNER *scan);
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define npc_attset(p)       ((p)->attset)
#define npc_attcnt(p)       ((p)->attcnt)
#define npc_valcnt(p,a)     ((p)->dvecs[a].valcnt)
#define npc_clsatt(p)       ((p)->clsatt)
#define npc_clsid(p)        ((p)->clsid)
#define npc_clscnt(p)       ((p)->clscnt)
#define npc_clsfrq(p,c)     ((p)->frqs[c])
#define npc_total(p)        ((p)->total)

#define npc_prior(p,c)      ((p)->priors[c])
#define npc_poss(p,c,a,v)   ((p)->dvecs[a].possds[c].probs[v])
#define npc_post(p,c)       ((p)->posts[c])

#endif
