/*----------------------------------------------------------------------
  File    : npx.c
  Contents: naive possibilistic classifier execution
  Author  : Christian Borgelt
  History : 2001.02.07 file created from file nbc.c
            2001.07.17 adapted to modified module scan
            2002.09.18 bug concerning missing target fixed
            2003.02.02 bug in alignment in connection with -d fixed
            2003.04.23 missing AS_MARKED added for second reading
            2003.08.16 slight changes in error message output
            2007.02.14 adapted to modified module attset
            2011.12.12 adapted to modified scanner functions
            2011.12.15 processing without table reading improved
            2013.08.09 adapted to definitions ATTID, VALID, TPLID etc.
            2013.08.20 output format changed to significant digits
            2014.10.24 changed from LGPL license to MIT license
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#ifndef AS_READ
#define AS_READ
#endif
#ifndef AS_WRITE
#define AS_WRITE
#endif
#ifndef AS_PARSE
#define AS_PARSE
#endif
#include "attset.h"
#ifndef TAB_READ
#define TAB_READ
#endif
#ifndef TAB_WRITE
#define TAB_WRITE
#endif
#include "table.h"
#ifndef NPC_PARSE
#define NPC_PARSE
#endif
#include "nposs.h"
#include "error.h"
#ifdef STORAGE
#include "storage.h"
#endif

#if ATTID_MAX < VALID_MAX
#error "npx requires ATTID_MAX >= VALID_MAX"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME     "npx"
#define DESCRIPTION "naive possibilistic classifier execution"
#define VERSION     "version 2.1 (2014.10.30)         " \
                    "(c) 1999-2014   Christian Borgelt"

/* --- error codes --- */
/* error codes 0 to -5 defined in attset.h */
#define E_OPTION    (-6)        /* unknown option */
#define E_OPTARG    (-7)        /* missing option argument */
#define E_ARGCNT    (-8)        /* wrong number of arguments */
#define E_PARSE     (-9)        /* parse error on classifier file */
#define E_OUTPUT   (-10)        /* class in input or write output */

#define SEC_SINCE(t)  ((double)(clock()-(t)) /(double)CLOCKS_PER_SEC)

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- classification result info.--- */
  ATT   *att;                   /* class attribute */
  VALID  cnt;                   /* number of classes */
  VALID  class;                 /* class (classification result) */
  CCHAR *col_cls;               /* name   of classification column */
  int    cwd_cls;               /* width  of classification column */
  double poss;                  /* degree of possibility of result */
  CCHAR *col_poss;              /* name   of possibility column */
  int    cwd_poss;              /* width  of possibility column */
  int    dig_poss;              /* digits of possibility column */
  int    all;                   /* whether to show all possibilities */
  int    cwd_all;               /* width  of possibility columns */
  int    err;                   /* classification error indicator */
} RESULT;                       /* (classification result info.) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *errmsgs[] = {   /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file %s",
  /* E_FREAD    -3 */  "read error on file %s",
  /* E_FWRITE   -4 */  "write error on file %s",
  /* E_STDIN    -5 */  "double assignment of standard input",
  /* E_OPTION   -6 */  "unknown option -%c",
  /* E_OPTARG   -7 */  "missing option argument",
  /* E_ARGCNT   -8 */  "wrong number of arguments",
  /* E_PARSE    -9 */  "parse error(s) on file %s",
  /* E_OUTPUT  -10 */  "must have class as input or write output",
  /*           -11 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static CCHAR    *prgname;       /* program name for error messages */
static SCANNER  *scan   = NULL; /* scanner (for poss. classifier) */
static TABREAD  *tread  = NULL; /* table reader */
static TABWRITE *twrite = NULL; /* table writer */
static ATTSET   *attset = NULL; /* attribute set */
static TABLE    *table  = NULL; /* data table */
static NPC      *npc    = NULL; /* naive possibilistic classifier */
static RESULT   res     = {     /* classification result information */
  NULL, 0,                      /* class attribute and # of classes */
  0,  "npc", 0,                 /* data for classification column */
  0.0, NULL, 0, 3,              /* data for possibility    column */
  0, 0,                         /* flag for all possibility degrees */
  0 };                          /* classification error indicator */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

#ifndef NDEBUG                  /* if debug version */
  #undef  CLEANUP               /* clean up memory and close files */
  #define CLEANUP \
  if (npc)    npc_delete(npc, 0);    \
  if (attset) as_delete(attset);     \
  if (tread)  trd_delete(tread,  1); \
  if (twrite) twr_delete(twrite, 1); \
  if (scan)   scn_delete(scan,   1);
#endif

GENERROR(error, exit)           /* generic error reporting function */

/*--------------------------------------------------------------------*/

static void classify (void)
{                               /* --- classify the current tuple */
  VALID c;                      /* buffer for actual class */

  assert(npc);                  /* execute the classifier */
  res.class = npc_exec(npc, NULL, &res.poss);
  c = att_inst(res.att)->n;     /* compute the error indicator */
  res.err = (!isnull(c) && (res.class != c)) ? 1 : 0;
}  /* classify() */

/*--------------------------------------------------------------------*/

static void infout (ATTSET *set, TABWRITE *twrite, int mode)
{                               /* --- write additional information */
  int   n, k;                   /* character counters */
  VALID c;                      /* loop variable for classes */

  assert(set && twrite);        /* check the function arguments */
  if (mode & AS_ATT) {          /* if to write the header */
    twr_puts(twrite, res.col_cls);  /* write class column name */
    if ((mode & AS_ALIGN)       /* if to align the column */
    && ((mode & AS_WEIGHT) || res.col_poss || res.all)){
      n = (int)strlen(res.col_cls);
      k = att_valwd(res.att, 0);
      res.cwd_cls = k = ((mode & AS_ALNHDR) && (n > k)) ? n : k;
      if (k > n) twr_pad(twrite, (size_t)(k-n));
    }                           /* compute width of class column */
    if (res.col_poss) {         /* if to write a class possibility */
      twr_fldsep(twrite);       /* write a field separator and */
      twr_puts(twrite, res.col_poss);       /* the column name */
      if ((mode & AS_ALIGN)     /* if to align the column */
      && ((mode & AS_WEIGHT) || res.all)) {
        n = (int)strlen(res.col_poss);
        k = res.dig_poss +3;    /* compute width of poss. column */
        res.cwd_poss = k = ((mode & AS_ALNHDR) && (n > k)) ? n : k;
        if (k > n) twr_pad(twrite, (size_t)(k-n));
      }                         /* pad with blanks if requested */
    }
    if (res.all) {              /* if to write all possibilities */
      if (mode & AS_ALIGN) {    /* if to align the columns */
        n = att_valwd(res.att, 0);
        k = res.dig_poss +3;    /* compute width of poss. columns */
        res.cwd_all = ((mode & AS_ALNHDR) && (n > k)) ? n : k;
      }
      for (c = 0; c < res.cnt; c++) {
        twr_fldsep(twrite);     /* traverse the target values */
        n = twr_printf(twrite, "%s", att_valname(res.att, c));
        if (res.cwd_all > n) twr_pad(twrite, (size_t)(res.cwd_all-n));
      }                         /* print the value name and */
    } }                         /* pad with blanks if requested */
  else {                        /* if to write a normal record */
    n = twr_printf(twrite, "%s", att_valname(res.att, res.class));
    if (res.cwd_cls > n) twr_pad(twrite, (size_t)(res.cwd_cls-n));
    if (res.col_poss) {         /* if to write a class possibility */
      twr_fldsep(twrite);       /* write separator and possibility */
      n = twr_printf(twrite, "%.*g", res.dig_poss, res.poss);
      if (res.cwd_poss > n) twr_pad(twrite, (size_t)(res.cwd_poss-n));
    }                           /* if to align, pad with blanks */
    if (res.all) {              /* if to write all possibilities */
      for (c = 0; c < res.cnt; c++) {
        twr_fldsep(twrite);     /* traverse the target values */
        n = twr_printf(twrite, "%.*g", res.dig_poss, npc_post(npc,c));
        if (res.cwd_all > n) twr_pad(twrite, (size_t)(res.cwd_all-n));
      }                         /* print the possibility and */
    }                           /* pad with blanks if requested */
  }
}  /* infout() */

/*--------------------------------------------------------------------*/

int main (int argc, char* argv[])
{                               /* --- main function */
  int     i, k = 0;             /* loop variables, buffer */
  char    *s;                   /* to traverse options */
  CCHAR   **optarg = NULL;      /* option argument */
  CCHAR   *fn_hdr  = NULL;      /* name of table header file */
  CCHAR   *fn_tab  = NULL;      /* name of table file */
  CCHAR   *fn_npc  = NULL;      /* name of classifier file */
  CCHAR   *fn_out  = NULL;      /* name of output file */
  CCHAR   *recseps = NULL;      /* record     separators */
  CCHAR   *fldseps = NULL;      /* field      separators */
  CCHAR   *blanks  = NULL;      /* blank      characters */
  CCHAR   *nullchs = NULL;      /* null value characters */
  CCHAR   *comment = NULL;      /* comment    characters */
  int     mode     = AS_ATT|AS_MARKED; /* table file read  mode */
  int     mout     = AS_ATT;           /* table file write mode */
  double  errs     = 0.0;       /* number of misclassifications */
  ATTID   m;                    /* number of attributes */
  TPLID   n, r;                 /* number of data tuples */
  double  w, u;                 /* weight of data tuples, buffer */
  clock_t t;                    /* timer for measurements */

  prgname = argv[0];            /* get program name for error msgs. */

  /* --- print startup/usage message --- */
  if (argc > 1) {               /* if arguments are given */
    fprintf(stderr, "%s - %s\n", argv[0], DESCRIPTION);
    fprintf(stderr, VERSION); } /* print a startup message */
  else {                        /* if no argument given */
    printf("usage: %s [options] npcfile "
                     "[-d|-h hdrfile] tabfile [outfile]\n", argv[0]);
    printf("naive possibilistic classifier execution\n");
    printf("%s\n", VERSION);
    printf("-p#      prediction/classification field name   "
                    "(default: \"%s\")\n", res.col_cls);
    printf("-c#      confidence/possibility field name      "
                    "(default: no confidence)\n");
    printf("-o#      significant digits for confidence      "
                    "(default: %d)\n", res.dig_poss);
    printf("-x       print extended confidence information  "
                    "(all poss. degrees)\n");
    printf("-a       align fields in output table           "
                    "(default: single separator)\n");
    printf("-w       do not write field names to the output file\n");
    printf("-r#      record  separators                     "
                    "(default: \"\\n\")\n");
    printf("-f#      field   separators                     "
                    "(default: \" \\t,\")\n");
    printf("-b#      blank   characters                     "
                    "(default: \" \\t\\r\")\n");
    printf("-u#      null value characters                  "
                    "(default: \"?*\")\n");
    printf("-C#      comment characters                     "
                    "(default: \"#\")\n");
    printf("-n       number of tuple occurrences in last field\n");
    printf("npcfile  file containing classifier description\n");
    printf("-d       use default table header "
                    "(field names = field numbers)\n");
    printf("-h       read table header (field names) from hdrfile\n");
    printf("hdrfile  file containing table header (field names)\n");
    printf("tabfile  table file to read "
                    "(field names in first record)\n");
    printf("outfile  file to write output table to (optional)\n");
    return 0;                   /* print a usage message */
  }                             /* and abort the program */

  /* --- evaluate arguments --- */
  for (i = 1; i < argc; i++) {  /* traverse arguments */
    s = argv[i];                /* get option argument */
    if (optarg) { *optarg = s; optarg = NULL; continue; }
    if ((*s == '-') && *++s) {  /* -- if argument is an option */
      while (*s) {              /* traverse options */
        switch (*s++) {         /* evaluate option */
          case 'p': optarg  = &res.col_cls;  break;
          case 'c': optarg  = &res.col_poss; break;
          case 'o': res.dig_poss = (int)strtol(s, &s, 0); break;
          case 'x': res.all = 1;             break;
          case 'n': mode   |= AS_WEIGHT;
                    mout   |= AS_WEIGHT;     break;
          case 'a': mout   |= AS_ALIGN;      break;
          case 'w': mout   &= ~AS_ATT;       break;
          case 'r': optarg  = &recseps;      break;
          case 'f': optarg  = &fldseps;      break;
          case 'b': optarg  = &blanks;       break;
          case 'u': optarg  = &nullchs;      break;
          case 'C': optarg  = &comment;      break;
          case 'd': mode   |= AS_DFLT;       break;
          case 'h': optarg  = &fn_hdr;       break;
          default : error(E_OPTION, *--s);   break;
        }                       /* set option variables */
        if (!*s) break;         /* if at end of string, abort loop */
        if (optarg) { *optarg = s; optarg = NULL; break; }
      } }                       /* get option argument */
    else {                      /* if argument is no option */
      switch (k++) {            /* evaluate non-option */
        case  0: fn_npc = s;      break;
        case  1: fn_tab = s;      break;
        case  2: fn_out = s;      break;
        default: error(E_ARGCNT); break;
      }                         /* note filenames */
    }
  }
  if (optarg) error(E_OPTARG);  /* check the option argument */
  if ((k < 2) || (k > 3))       /* and the number of arguments */
    error(E_ARGCNT);            /* (output file is optional) */
  if (fn_hdr && (strcmp(fn_hdr, "-") == 0))
    fn_hdr = "";                /* convert "-" to "" */
  i = ( fn_hdr && !*fn_hdr) ? 1 : 0;
  if  (!fn_npc || !*fn_npc) i++;
  if  (!fn_tab || !*fn_tab) i++;/* check assignments of stdin */
  if (i > 1) error(E_STDIN);    /* stdin must not be used twice */
  if ((mout & AS_ATT) && (mout & AS_ALIGN))
    mout |= AS_ALNHDR;          /* set align to header flag */
  if (fn_out) mout |= AS_MARKED|AS_INFO1|AS_RDORD;
  else        mout  = 0;        /* set up the table write mode */
  fputc('\n', stderr);          /* terminate the startup message */

  /* --- read naive possibilistic classifier --- */
  attset = as_create("domains", att_delete);
  if (!attset) error(E_NOMEM);  /* create an attribute set */
  scan = scn_create();          /* create a scanner */
  if (!scan)   error(E_NOMEM);  /* for the classifier file */
  t = clock();                  /* start timer, open intput file */
  if (scn_open(scan, NULL, fn_npc) != 0)
    error(E_FOPEN, scn_name(scan));
  fprintf(stderr, "reading %s ... ", scn_name(scan));
  if (as_parse(attset, scan, AT_ALL, 1) != 0)
    error(E_PARSE, scn_name(scan));  /* parse domain descriptions */
  npc = npc_parse(attset,scan); /* and the naive poss. classifier */
  if (!npc || !scn_eof(scan, 1)) error(E_PARSE, scn_name(scan));
  scn_delete(scan, 1);          /* delete the scanner and */
  scan = NULL;                  /* clear the scanner variable */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", as_attcnt(attset));
  fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));

  /* --- get class attribute --- */
  res.att = npc_clsatt(npc);    /* get the class attribute */
  res.cnt = att_valcnt(res.att);/* and its number of values */
  as_setmark(attset, 1);        /* mark all attributes */
  att_setmark(res.att, 0);      /* except the class attribute */

  /* --- read table header --- */
  tread = trd_create();         /* create a table reader and */
  if (!tread) error(E_NOMEM);   /* configure the characters */
  trd_allchs(tread, recseps, fldseps, blanks, nullchs, comment);
  if (fn_hdr) {                 /* if a header file is given */
    t = clock();                /* start timer, open input file */
    if (trd_open(tread, NULL, fn_hdr) != 0)
      error(E_FOPEN, trd_name(tread));
    fprintf(stderr, "reading %s ... ", trd_name(tread));
    k = as_read(attset, tread, (mode & ~AS_DFLT) | AS_ATT);
    if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
    trd_close(tread);           /* read table header, close file */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s)]", as_attcnt(attset));
    fprintf(stderr, " done [%.2fs].\n", SEC_SINCE(t));
    mode &= ~(AS_ATT|AS_DFLT);  /* print a success message and */
  }                             /* remove the attribute flag */

  /* --- process table body --- */
  t = clock();                  /* start timer, open input file */
  if (trd_open(tread, NULL, fn_tab) != 0)
    error(E_FOPEN, trd_name(tread));
  fprintf(stderr, "reading %s ... ", trd_name(tread));
  if (mout & AS_ALIGN) {        /* if to align the output columns */
    table = tab_create("table", attset, tpl_delete);
    if (!table) error(E_NOMEM); /* read the data table */
    k = tab_read(table, tread, mode);
    if (k < 0) error(-k, tab_errmsg(table, NULL, 0));
    trd_delete(tread, 1);       /* delete the table reader */
    tread = NULL;               /* and clear the variable */
    m = tab_attcnt(table);      /* get the number of attributes */
    n = tab_tplcnt(table);      /* and the number of data tuples */
    w = tab_tplwgt(table);      /* and print a success message */
    fprintf(stderr, "[%"ATTID_FMT" attribute(s),", m);
    fprintf(stderr, " %"TPLID_FMT, n);
    if (w != (double)n) fprintf(stderr, "/%g", w);
    fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));
    twrite = twr_create();      /* create a table writer and */
    if (!twrite) error(E_NOMEM);/* configure the characters */
    twr_xchars(twrite, recseps, fldseps, blanks, nullchs);
    t = clock();                /* start timer, open output file */
    if (twr_open(twrite, NULL, fn_out) != 0)
      error(E_FOPEN, twr_name(twrite));
    fprintf(stderr, "writing %s ... ", twr_name(twrite));
    if ((mout & AS_ATT)         /* write a table header */
    &&  (as_write(attset, twrite, mout, infout) != 0))
      error(E_FWRITE, twr_name(twrite));
    mout = AS_INST | (mout & ~AS_ATT);
    m += (res.col_poss ? 2 : 1) +(res.all ? res.cnt : 0);
    for (r = 0; r < n; r++) {   /* traverse the tuples */
      tpl_toas(tab_tpl(table, r));
      classify();               /* classify the current tuple */
      u = as_getwgt(attset);    /* get the tuple weight and */
      errs += res.err *u;       /* count the misclassifications */
      if (as_write(attset, twrite, mout, infout) != 0)
        error(E_FWRITE, twr_name(twrite));
    } }                         /* write the current tuple */
  else {                        /* if to process tuples directly */
    k = as_read(attset, tread, mode); /* read/generate table header */
    if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
    if (!fn_out && (att_getmark(res.att) < 0))
      error(E_OUTPUT);          /* check for output to produce */
    if (fn_out) {               /* if to write an output file */
      twrite = twr_create();    /* create a table writer and */
      if (!twrite) error(E_NOMEM);   /* configure characters */
      twr_xchars(twrite, recseps, fldseps, blanks, nullchs);
      t = clock();              /* start timer, open output file */
      if (twr_open(twrite, NULL, fn_out) != 0)
        error(E_FOPEN, twr_name(twrite));
      if ((mout & AS_ATT)       /* write a table header */
      &&  (as_write(attset, twrite, mout, infout) != 0))
        error(E_FWRITE, twr_name(twrite));
      mout = AS_INST | (mout & ~AS_ATT);
    }                           /* remove the attribute flag */
    i = mode; mode = (mode & ~(AS_DFLT|AS_ATT)) | AS_INST;
    if (!(i & AS_DFLT))         /* if not done yet, read first tuple */
      k = as_read(attset, tread, mode);
    for (w = 0, n = 0; k == 0; n++) {
      classify();               /* classify the current tuple */
      w    += u = as_getwgt(attset);   /* sum the tuple weights */
      errs += res.err *u;       /* count the misclassifications */
      if (twrite                /* write the current tuple */
      && (as_write(attset, twrite, mout, infout) != 0))
        error(E_FWRITE, twr_name(twrite));
      k = as_read(attset, tread, mode);
    }                           /* try to read the next tuple */
    if (k < 0) error(-k, as_errmsg(attset, NULL, 0));
    trd_delete(tread, 1);       /* delete the table reader */
    tread = NULL;               /* and clear the variable */
    m = as_attcnt(attset);      /* get the number of attributes */
  }
  if (twrite) {                 /* if an output file was written */
    if (twr_close(twrite) != 0) error(E_FWRITE, twr_name(twrite));
    twr_delete(twrite, 1);      /* close the output file and */
    twrite = NULL;              /* delete the table writer */
  }                             /* print a success message */
  fprintf(stderr, "[%"ATTID_FMT" attribute(s),", m);
  fprintf(stderr, " %"TPLID_FMT, n);
  if (w != (double)n) fprintf(stderr, "/%g", w);
  fprintf(stderr, " tuple(s)] done [%.2fs].\n", SEC_SINCE(t));

  /* --- print error statistics --- */
  if (att_getmark(res.att) >= 0) { /* if the target is present */
    fprintf(stderr, "%g error(s) ", errs);
    fprintf(stderr, "(%.2f%%)\n", (w > 0) ? 100*(errs/w) : 0);
  }                             /* print number of misclassifications */

  /* --- clean up --- */
  CLEANUP;                      /* clean up memory and close files */
  SHOWMEM;                      /* show (final) memory usage */
  return 0;                     /* return 'ok' */
}  /* main() */
